//  Module: CODEGEN.CPP

//                                                   September 20, 1993
//                                                   Fairfield, Iowa

//          Aerosoft  (R)  Register Wizard            Version 1.0
//          Copyright (c)  Aerosoft 1993     All rights reserved.

//          This software source code is FREEWARE.  You may be use the
//          source code or redistribute the source code free of charge.
//          However, you cannot sell this source code or any alteration of
//          the source code.

//  This module implements four functions that each perform code generation
//  for one of the four files:

//      CHNNLDEF.H, REGISTER.H, REGISTER.CPP, APP.CPP

//  See file BCWIZ.MAN for an explanation of these files.

#include    <stdio.h>
#include    <string.h>
#include    <process.h>

#include    "CODEGEN.H"
        
void GenCHNNLDEF_H(ChannelSpecs &oChannelSpecs)
{
    FILE    *fid;
    
    if ((fid=fopen(CHNNLDEF_H_FILE, "wt"))==NULL)
    {
        printf("*** Execution error: Unable to open %s.\n", CHNNLDEF_H_FILE); 
        exit(0);
    }
  
    fprintf(fid, "#ifndef _chnnldef_h_\n");
    fprintf(fid, "#define _chnnldef_h_\n\n");
    fprintf(fid, "//  Module: CHNNLDEF.H\n\n"); 

    ChannelSpec *opSpec = oChannelSpecs.oChannelSpec;   
    for (int i=0; i<oChannelSpecs.iNumSpecs; i++,opSpec++)
    {
        fprintf(fid, "class %s {\n", opSpec->cpFuncName);
        fprintf(fid, "public:\n");
        fprintf(fid, "virtual %s Serve%s %s=0;\n", opSpec->cpReturnType,
                                                   opSpec->cpFuncName,
                                                   opSpec->cpProtoList);
        fprintf(fid, "};\n\n");
    }

    fprintf(fid, "#endif\n");   
    fclose(fid);
}

void GenREGISTER_CPP(ChannelSpecs &oChannelSpecs)
{
    FILE    *fid;
    
    if ((fid=fopen(REGISTER_CPP_FILE, "wt"))==NULL)
    {
        printf("*** Execution error: Unable to open %s.\n", REGISTER_CPP_FILE); 
        exit(0);
    }
  
    fprintf(fid, "//  Module: REGISTER.CPP\n\n");
    fprintf(fid, "#include <stdio.h>\n");
    fprintf(fid, "#include \"%s\"\n", REGISTER_H_FILE);
    fprintf(fid, "#include \"%s\"\n\n", CHNNLDEF_H_FILE);
    fprintf(fid, "    Register oRegister;\n\n");
    
    ChannelSpec *opSpec = oChannelSpecs.oChannelSpec;   
    for (int i=0; i<oChannelSpecs.iNumSpecs; i++,opSpec++)
    {
        fprintf(fid, "int Register::Connect (%s *op) {\n",
                                               opSpec->cpFuncName);
        fprintf(fid, "    if (op%sChannel == NULL) {\n", opSpec->cpFuncName);
        fprintf(fid, "        op%sChannel = op;\n", opSpec->cpFuncName);
        fprintf(fid, "        return 1;\n");
        fprintf(fid, "    } else {\n");
        fprintf(fid, "        return 0;\n");
        fprintf(fid, "    }\n");
        fprintf(fid, "}\n");

        fprintf(fid, "int Register::Disconnect (%s *op) {\n",
                                               opSpec->cpFuncName);
        fprintf(fid, "    if (op%sChannel == op) {\n", opSpec->cpFuncName);
        fprintf(fid, "        op%sChannel = NULL;\n", opSpec->cpFuncName);
        fprintf(fid, "        return 1;\n");
        fprintf(fid, "    } else {\n");
        fprintf(fid, "        return 0;\n");
        fprintf(fid, "    }\n");
        fprintf(fid, "}\n");

        fprintf(fid, "%s Register::Execute%s %s {\n",
                           opSpec->cpReturnType,
                           opSpec->cpFuncName, opSpec->cpProtoList);

        fprintf(fid, "    if (op%sChannel == NULL)\n", opSpec->cpFuncName);
        if (stricmp("void", opSpec->cpReturnType)==0)
            fprintf(fid, "        return;\n");
        else
            fprintf(fid, "        return (%s)%s;\n\n",
                  opSpec->cpReturnType, opSpec->cpReturnDefaultValue);
                  
        if (stricmp("void", opSpec->cpReturnType)==0)
            fprintf(fid, "    ((%s *)op%sChannel)->Serve%s%s;\n",
                  opSpec->cpFuncName,
                  opSpec->cpFuncName, opSpec->cpFuncName, opSpec->cpArgList);
        else
            fprintf(fid, "    return ((%s *)op%sChannel)->Serve%s%s;\n",
                  opSpec->cpFuncName,
                  opSpec->cpFuncName, opSpec->cpFuncName, opSpec->cpArgList);

        fprintf(fid, "}\n");
    }

    fclose(fid);
}

void GenREGISTER_H(ChannelSpecs &oChannelSpecs)
{
    FILE    *fid;
    
    if ((fid=fopen(REGISTER_H_FILE, "wt"))==NULL)
    {
        printf("*** Execution error: Unable to open %s.\n", REGISTER_H_FILE); 
        exit(0);
    }
  
    fprintf(fid, "#ifndef _register_h_\n");
    fprintf(fid, "#define _register_h_\n\n");
    
    fprintf(fid, "//  Module: REGISTER.H\n\n");
    fprintf(fid, "#include \"%s\"\n\n", CHNNLDEF_H_FILE);
    fprintf(fid, "class Register {\n");
    
    ChannelSpec *opSpec = oChannelSpecs.oChannelSpec;   
    for (int i=0; i<oChannelSpecs.iNumSpecs; i++,opSpec++)
    {
        fprintf(fid, "    void *op%sChannel;\n", opSpec->cpFuncName);
    }

    fprintf(fid, "public:\n");    
    opSpec = oChannelSpecs.oChannelSpec;    
    for (i=0; i<oChannelSpecs.iNumSpecs; i++,opSpec++)
    {
        fprintf(fid, "    int Connect (%s *op);\n", opSpec->cpFuncName);
        fprintf(fid, "    int Disconnect (%s *op);\n", opSpec->cpFuncName);
        fprintf(fid, "    int ChannelStatus (%s *)\n", opSpec->cpFuncName);
        fprintf(fid, "    {return op%sChannel!=NULL;}\n", opSpec->cpFuncName);
        fprintf(fid, "    %s Execute%s %s;\n",
                                opSpec->cpReturnType,
                                opSpec->cpFuncName, opSpec->cpProtoList);
    }

    fprintf(fid, "};\n\n");
    fprintf(fid, "    extern Register    oRegister;\n\n");
    fprintf(fid, "#endif\n");
    fclose(fid);
}


void GenAPP_CPP(ChannelSpecs &oChannelSpecs)
{
    FILE    *fid;
    
    if ((fid=fopen(APP_CPP_FILE, "wt"))==NULL)
    {
        printf("*** Execution error: Unable to open %s.\n", APP_CPP_FILE); 
        exit(0);
    }
  
    fprintf(fid, "//  Module: APP.CPP\n\n");
    fprintf(fid, "//  Cut and paste the contents of this module to create\n");
    fprintf(fid, "//  application code.  There are two sections: Client\n");
    fprintf(fid, "//  calls and Server methods.\n\n");
    fprintf(fid, "#include \"%s\"\n\n", CHNNLDEF_H_FILE);

    fprintf(fid, "//  Client calls --\n\n");    
    ChannelSpec *opSpec = oChannelSpecs.oChannelSpec;   
    for (int i=0; i<oChannelSpecs.iNumSpecs; i++,opSpec++)
    {
        fprintf(fid, "    oRegister.Execute%s%s;\n",
                               opSpec->cpFuncName, opSpec->cpArgList);
    }
    
    fprintf(fid, "\n//  Server methods --\n\n");    
    opSpec = oChannelSpecs.oChannelSpec;    
    for (i=0; i<oChannelSpecs.iNumSpecs; i++,opSpec++)
    {
        fprintf(fid, "class Server%d : public %s {\n",
                               i+1, opSpec->cpFuncName);
        fprintf(fid, "public:\n");
        fprintf(fid, "    void Serve%s %s;\n",
                               opSpec->cpFuncName, opSpec->cpProtoList);
        fprintf(fid, "};\n");
        fprintf(fid, "void Server%d::Serve%s %s {}\n\n",
                          i+1, opSpec->cpFuncName, opSpec->cpProtoList);
    }
    
    fclose(fid);
}
